"use strict";
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _RedisCommandsQueue_instances, _a, _RedisCommandsQueue_flushWaitingForReply, _RedisCommandsQueue_flushWaitingToBeSent;
Object.defineProperty(exports, "__esModule", { value: true });
const LinkedList = require("yallist");
const encoder_1 = require("../RESP/encoder");
const decoder_1 = require("../RESP/decoder");
const pub_sub_1 = require("./pub-sub");
const errors_1 = require("../errors");
const PONG = Buffer.from('pong');
const RESP2_PUSH_FLAGS = {
    ...decoder_1.PUSH_FLAGS,
    [decoder_1.RESP_TYPES.SIMPLE_STRING]: Buffer
};
class RedisCommandsQueue {
    get isPubSubActive() {
        return this._pubSub.isActive;
    }
    constructor(respVersion, maxLength, onShardedChannelMoved) {
        _RedisCommandsQueue_instances.add(this);
        Object.defineProperty(this, "_maxLength", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_waitingToBeSent", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new LinkedList()
        });
        Object.defineProperty(this, "_waitingForReply", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new LinkedList()
        });
        Object.defineProperty(this, "_onShardedChannelMoved", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_pubSub", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new pub_sub_1.PubSub()
        });
        Object.defineProperty(this, "_chainInExecution", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "decoder", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.decoder = this._initiateDecoder(respVersion);
        this._maxLength = maxLength;
        this._onShardedChannelMoved = onShardedChannelMoved;
    }
    _initiateDecoder(respVersion) {
        return respVersion === 3 ?
            this._initiateResp3Decoder() :
            this._initiateResp2Decoder();
    }
    _onReply(reply) {
        this._waitingForReply.shift().resolve(reply);
    }
    _onErrorReply(err) {
        this._waitingForReply.shift().reject(err);
    }
    _onPush(push) {
        // TODO: type
        if (this._pubSub.handleMessageReply(push))
            return true;
        const isShardedUnsubscribe = pub_sub_1.PubSub.isShardedUnsubscribe(push);
        if (isShardedUnsubscribe && !this._waitingForReply.length) {
            const channel = push[1].toString();
            this._onShardedChannelMoved(channel, this._pubSub.removeShardedListeners(channel));
            return true;
        }
        else if (isShardedUnsubscribe || pub_sub_1.PubSub.isStatusReply(push)) {
            const head = this._waitingForReply.head.value;
            if ((Number.isNaN(head.channelsCounter) && push[2] === 0) ||
                --head.channelsCounter === 0) {
                this._waitingForReply.shift().resolve();
            }
            return true;
        }
    }
    _getFlags() {
        return this._waitingForReply.head.value.flags ?? {};
    }
    _initiateResp3Decoder() {
        return new decoder_1.Decoder({
            onReply: reply => this._onReply(reply),
            onErrorReply: err => this._onErrorReply(err),
            onPush: push => {
                if (!this._onPush(push)) {
                }
            },
            getFlags: () => this._getFlags()
        });
    }
    _initiateResp2Decoder() {
        return new decoder_1.Decoder({
            onReply: reply => {
                if (this._pubSub.isActive && Array.isArray(reply)) {
                    if (this._onPush(reply))
                        return;
                    if (PONG.equals(reply[0])) {
                        const { resolve, flags } = this._waitingForReply.shift(), buffer = (reply[1].length === 0 ? reply[0] : reply[1]);
                        resolve(flags?.[decoder_1.RESP_TYPES.SIMPLE_STRING] === Buffer ? buffer : buffer.toString());
                        return;
                    }
                }
                this._onReply(reply);
            },
            onErrorReply: err => this._onErrorReply(err),
            // PUSH type does not exist in RESP2
            // PubSub is handled in onReply  
            // @ts-expect-error
            onPush: undefined,
            getFlags: () => {
                // PubSub push is an Array in RESP2
                return this._pubSub.isActive ?
                    RESP2_PUSH_FLAGS :
                    this._getFlags();
            }
        });
    }
    addCommand(args, options) {
        if (this._maxLength && this._waitingToBeSent.length + this._waitingForReply.length >= this._maxLength) {
            return Promise.reject(new Error('The queue is full'));
        }
        else if (options?.signal?.aborted) {
            return Promise.reject(new errors_1.AbortError());
        }
        return new Promise((resolve, reject) => {
            const node = new LinkedList.Node({
                args,
                chainId: options?.chainId,
                flags: options?.flags,
                resolve,
                reject
            });
            if (options?.signal) {
                const listener = () => {
                    this._waitingToBeSent.removeNode(node);
                    node.value.reject(new errors_1.AbortError());
                };
                node.value.removeAbortListener = () => options.signal?.removeEventListener('abort', listener);
                options.signal.addEventListener('abort', listener, { once: true });
            }
            if (options?.asap) {
                this._waitingToBeSent.unshiftNode(node);
            }
            else {
                this._waitingToBeSent.pushNode(node);
            }
        });
    }
    subscribe(type, channels, listener, returnBuffers) {
        return this._pushPubSubCommand(this._pubSub.subscribe(type, channels, listener, returnBuffers));
    }
    unsubscribe(type, channels, listener, returnBuffers) {
        return this._pushPubSubCommand(this._pubSub.unsubscribe(type, channels, listener, returnBuffers));
    }
    resubscribe() {
        const commands = this._pubSub.resubscribe();
        if (!commands.length)
            return;
        return Promise.all(commands.map(command => this._pushPubSubCommand(command)));
    }
    extendPubSubChannelListeners(type, channel, listeners) {
        return this._pushPubSubCommand(this._pubSub.extendChannelListeners(type, channel, listeners));
    }
    extendPubSubListeners(type, listeners) {
        return this._pushPubSubCommand(this._pubSub.extendTypeListeners(type, listeners));
    }
    getPubSubListeners(type) {
        return this._pubSub.getTypeListeners(type);
    }
    _pushPubSubCommand(command) {
        if (command === undefined)
            return;
        return new Promise((resolve, reject) => {
            this._waitingToBeSent.push({
                args: command.args,
                channelsCounter: command.channelsCounter,
                flags: decoder_1.PUSH_FLAGS,
                resolve: () => {
                    command.resolve();
                    resolve();
                },
                reject: err => {
                    command.reject?.();
                    reject(err);
                }
            });
        });
    }
    getCommandToSend() {
        const toSend = this._waitingToBeSent.shift();
        if (!toSend)
            return;
        let encoded;
        try {
            encoded = (0, encoder_1.default)(toSend.args);
        }
        catch (err) {
            toSend.reject(err);
            return;
        }
        // TODO
        // reuse `toSend`
        toSend.args = undefined;
        if (toSend.removeAbortListener) {
            toSend.removeAbortListener();
            toSend.removeAbortListener = undefined;
        }
        this._waitingForReply.push(toSend);
        this._chainInExecution = toSend.chainId;
        return encoded;
    }
    flushWaitingForReply(err) {
        this.decoder.reset();
        this._pubSub.reset();
        __classPrivateFieldGet(this, _RedisCommandsQueue_instances, "m", _RedisCommandsQueue_flushWaitingForReply).call(this, err);
        if (!this._chainInExecution)
            return;
        while (this._waitingToBeSent.head?.value.chainId === this._chainInExecution) {
            __classPrivateFieldGet(RedisCommandsQueue, _a, "m", _RedisCommandsQueue_flushWaitingToBeSent).call(RedisCommandsQueue, this._waitingToBeSent.shift(), err);
        }
        this._chainInExecution = undefined;
    }
    flushAll(err) {
        this.decoder.reset();
        this._pubSub.reset();
        __classPrivateFieldGet(this, _RedisCommandsQueue_instances, "m", _RedisCommandsQueue_flushWaitingForReply).call(this, err);
        while (this._waitingToBeSent.head) {
            __classPrivateFieldGet(RedisCommandsQueue, _a, "m", _RedisCommandsQueue_flushWaitingToBeSent).call(RedisCommandsQueue, this._waitingToBeSent.shift(), err);
        }
    }
}
exports.default = RedisCommandsQueue;
_a = RedisCommandsQueue, _RedisCommandsQueue_instances = new WeakSet(), _RedisCommandsQueue_flushWaitingForReply = function _RedisCommandsQueue_flushWaitingForReply(err) {
    while (this._waitingForReply.head) {
        this._waitingForReply.shift().reject(err);
    }
}, _RedisCommandsQueue_flushWaitingToBeSent = function _RedisCommandsQueue_flushWaitingToBeSent(command, err) {
    command.removeAbortListener?.();
    command.reject(err);
};
