/// <reference types="node" />
/// <reference types="node" />
import COMMANDS from '../commands';
import { RedisSocketOptions } from './socket';
import { QueueCommandOptions } from './commands-queue';
import { EventEmitter } from 'events';
import { PubSubType, PubSubListener, PubSubTypeListeners, ChannelListeners } from './pub-sub';
import { CommandArguments, CommandSignature, Flags, CommanderConfig, RedisFunctions, RedisModules, RedisScripts, ReplyUnion, RespVersions, RedisArgument } from '../RESP/types';
import { RedisClientMultiCommandType } from './multi-command';
import { Options as PoolOptions } from 'generic-pool';
import { ReplyWithFlags, CommandReply } from '../RESP/types';
import SCAN, { ScanOptions, ScanCommonOptions } from '../commands/SCAN';
export interface RedisClientOptions<M extends RedisModules = RedisModules, F extends RedisFunctions = RedisFunctions, S extends RedisScripts = RedisScripts, RESP extends RespVersions = RespVersions> extends CommanderConfig<M, F, S, RESP> {
    /**
     * `redis[s]://[[username][:password]@][host][:port][/db-number]`
     * See [`redis`](https://www.iana.org/assignments/uri-schemes/prov/redis) and [`rediss`](https://www.iana.org/assignments/uri-schemes/prov/rediss) IANA registration for more details
     */
    url?: string;
    /**
     * Socket connection properties
     */
    socket?: RedisSocketOptions;
    /**
     * ACL username ([see ACL guide](https://redis.io/topics/acl))
     */
    username?: string;
    /**
     * ACL password or the old "--requirepass" password
     */
    password?: string;
    /**
     * Client name ([see `CLIENT SETNAME`](https://redis.io/commands/client-setname))
     */
    name?: string;
    /**
     * Redis database number (see [`SELECT`](https://redis.io/commands/select) command)
     */
    database?: number;
    /**
     * Maximum length of the client's internal command queue
     */
    commandsQueueMaxLength?: number;
    /**
     * When `true`, commands are rejected when the client is reconnecting.
     * When `false`, commands are queued for execution after reconnection.
     */
    disableOfflineQueue?: boolean;
    /**
     * Connect in [`READONLY`](https://redis.io/commands/readonly) mode
     */
    readonly?: boolean;
    /**
     * TODO
     */
    legacyMode?: boolean;
    /**
     * TODO
     */
    isolationPoolOptions?: PoolOptions;
    /**
     * Send `PING` command at interval (in ms).
     * Useful with Redis deployments that do not use TCP Keep-Alive.
     */
    pingInterval?: number;
}
type WithCommands<RESP extends RespVersions, FLAGS extends Flags> = {
    [P in keyof typeof COMMANDS]: CommandSignature<(typeof COMMANDS)[P], RESP, FLAGS>;
};
type WithModules<M extends RedisModules, RESP extends RespVersions, FLAGS extends Flags> = {
    [P in keyof M]: {
        [C in keyof M[P]]: CommandSignature<M[P][C], RESP, FLAGS>;
    };
};
type WithFunctions<F extends RedisFunctions, RESP extends RespVersions, FLAGS extends Flags> = {
    [L in keyof F]: {
        [C in keyof F[L]]: CommandSignature<F[L][C], RESP, FLAGS>;
    };
};
type WithScripts<S extends RedisScripts, RESP extends RespVersions, FLAGS extends Flags> = {
    [P in keyof S]: CommandSignature<S[P], RESP, FLAGS>;
};
export type RedisClientType<M extends RedisModules = {}, F extends RedisFunctions = {}, S extends RedisScripts = {}, RESP extends RespVersions = 2, FLAGS extends Flags = {}> = (RedisClient<M, F, S, RESP, FLAGS> & WithCommands<RESP, FLAGS> & WithModules<M, RESP, FLAGS> & WithFunctions<F, RESP, FLAGS> & WithScripts<S, RESP, FLAGS>);
export interface ClientCommandOptions extends QueueCommandOptions {
    isolated?: boolean;
}
interface ScanIteratorOptions {
    cursor?: number;
}
export default class RedisClient<M extends RedisModules, F extends RedisFunctions, S extends RedisScripts, RESP extends RespVersions, FLAGS extends Flags> extends EventEmitter {
    private static _createCommand;
    private static _createModuleCommand;
    private static _createFunctionCommand;
    private static _createScriptCommand;
    static factory<M extends RedisModules = {}, F extends RedisFunctions = {}, S extends RedisScripts = {}, RESP extends RespVersions = 2>(config?: CommanderConfig<M, F, S, RESP>): (options?: Omit<RedisClientOptions, keyof Exclude<typeof config, undefined>>) => RedisClientType<M, F, S, RESP, {}>;
    static create<M extends RedisModules = {}, F extends RedisFunctions = {}, S extends RedisScripts = {}, RESP extends RespVersions = 2>(options?: RedisClientOptions<M, F, S, RESP>): RedisClientType<M, F, S, RESP, {}>;
    static parseURL(url: string): RedisClientOptions;
    self: this;
    private readonly _options?;
    private readonly _socket;
    private readonly _queue;
    private _isolationPool?;
    private _selectedDB;
    get options(): RedisClientOptions<M, F, S, RESP> | undefined;
    get isOpen(): boolean;
    get isReady(): boolean;
    get isPubSubActive(): boolean;
    constructor(options?: RedisClientOptions<M, F, S, RESP>);
    private _initiateOptions;
    private _initiateQueue;
    private _initiateSocket;
    private _pingTimer?;
    private _setPingTimer;
    withCommandOptions<T extends ClientCommandOptions>(options: T): RedisClientType<M, F, S, RESP, T["flags"] extends Flags ? T["flags"] : {}>;
    private _commandOptionsProxy;
    /**
     * Override the `flags` command option
     */
    withFlags<FLAGS extends Flags>(flags: FLAGS): RedisClientType<M, F, S, RESP, FLAGS extends Flags ? FLAGS : {}>;
    /**
     * Override the `asap` command option to `true`
     */
    asap(): RedisClientType<M, F, S, RESP, FLAGS>;
    /**
     * Override the `isolated` command option to `true`
     */
    isolated(): RedisClientType<M, F, S, RESP, FLAGS>;
    duplicate(overrides?: Partial<RedisClientOptions<M, F, S, RESP>>): RedisClientType<M, F, S, RESP, {}>;
    connect(): Promise<void>;
    sendCommand: <T = ReplyUnion>(args: CommandArguments, options?: ClientCommandOptions) => Promise<T>;
    _sendCommand<T = ReplyUnion>(args: CommandArguments, options?: ClientCommandOptions): Promise<T>;
    SELECT(db: number): Promise<void>;
    select: (db: number) => Promise<void>;
    private _pubSubCommand;
    SUBSCRIBE<T extends boolean = false>(channels: string | Array<string>, listener: PubSubListener<T>, bufferMode?: T): Promise<void>;
    subscribe: <T extends boolean = false>(channels: string | Array<string>, listener: PubSubListener<T>, bufferMode?: T | undefined) => Promise<void>;
    UNSUBSCRIBE<T extends boolean = false>(channels?: string | Array<string>, listener?: PubSubListener<T>, bufferMode?: T): Promise<void>;
    unsubscribe: <T extends boolean = false>(channels?: string | Array<string>, listener?: PubSubListener<T> | undefined, bufferMode?: T | undefined) => Promise<void>;
    PSUBSCRIBE<T extends boolean = false>(patterns: string | Array<string>, listener: PubSubListener<T>, bufferMode?: T): Promise<void>;
    pSubscribe: <T extends boolean = false>(patterns: string | Array<string>, listener: PubSubListener<T>, bufferMode?: T | undefined) => Promise<void>;
    PUNSUBSCRIBE<T extends boolean = false>(patterns?: string | Array<string>, listener?: PubSubListener<T>, bufferMode?: T): Promise<void>;
    pUnsubscribe: <T extends boolean = false>(patterns?: string | Array<string>, listener?: PubSubListener<T> | undefined, bufferMode?: T | undefined) => Promise<void>;
    SSUBSCRIBE<T extends boolean = false>(channels: string | Array<string>, listener: PubSubListener<T>, bufferMode?: T): Promise<void>;
    sSubscribe: <T extends boolean = false>(channels: string | Array<string>, listener: PubSubListener<T>, bufferMode?: T | undefined) => Promise<void>;
    SUNSUBSCRIBE<T extends boolean = false>(channels?: string | Array<string>, listener?: PubSubListener<T>, bufferMode?: T): Promise<void>;
    sUnsubscribe: <T extends boolean = false>(channels?: string | Array<string>, listener?: PubSubListener<T> | undefined, bufferMode?: T | undefined) => Promise<void>;
    getPubSubListeners(type: PubSubType): PubSubTypeListeners;
    extendPubSubChannelListeners(type: PubSubType, channel: string, listeners: ChannelListeners): Promise<void>;
    extendPubSubListeners(type: PubSubType, listeners: PubSubTypeListeners): Promise<void>;
    QUIT(): Promise<string>;
    quit: () => Promise<string>;
    _tick(force?: boolean): void;
    executeIsolated<T>(fn: (client: RedisClientType<M, F, S, RESP, FLAGS>) => T | Promise<T>): Promise<T>;
    private _addMultiCommands;
    MULTI(): RedisClientMultiCommandType<[], M, F, S, RESP, FLAGS>;
    multi: () => RedisClientMultiCommandType<[], M, F, S, RESP, FLAGS>;
    scanIterator(this: RedisClientType<M, F, S, RESP, FLAGS>, options?: ScanOptions & ScanIteratorOptions): AsyncIterable<ReplyWithFlags<CommandReply<typeof SCAN, RESP>['keys'], FLAGS>>;
    hScanIterator(this: RedisClientType<M, F, S, RESP, FLAGS>, key: RedisArgument, options?: ScanCommonOptions & ScanIteratorOptions): AsyncGenerator<{
        field: (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string) | (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string);
        value: (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string) | (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string);
    }[], void, unknown>;
    sScanIterator(this: RedisClientType<M, F, S, RESP, FLAGS>, key: RedisArgument, options?: ScanCommonOptions & ScanIteratorOptions): AsyncGenerator<((FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string) | (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string))[], void, unknown>;
    zScanIterator(this: RedisClientType<M, F, S, RESP, FLAGS>, key: RedisArgument, options?: ScanCommonOptions & ScanIteratorOptions): AsyncGenerator<{
        value: (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string) | (FLAGS[36] extends import("../RESP/types").Flag<infer T> ? ReplyWithFlags<Extract<string, T>, FLAGS> | ReplyWithFlags<Extract<Buffer, T>, FLAGS> : string);
        score: number;
    }[], void, unknown>;
    disconnect(): Promise<void>;
    private _destroyIsolationPool;
    ref(): void;
    unref(): void;
}
export {};
