"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const commands_1 = require("../commands");
const socket_1 = require("./socket");
const commands_queue_1 = require("./commands-queue");
const events_1 = require("events");
const commander_1 = require("../commander");
const errors_1 = require("../errors");
const url_1 = require("url");
const pub_sub_1 = require("./pub-sub");
const multi_command_1 = require("./multi-command");
const HELLO_1 = require("../commands/HELLO");
const generic_pool_1 = require("generic-pool");
class RedisClient extends events_1.EventEmitter {
    static _createCommand(command, resp) {
        const transformReply = (0, commander_1.getTransformReply)(command, resp);
        return async function () {
            const args = command.transformArguments.apply(undefined, arguments), reply = await this._sendCommand(args, this.commandOptions);
            return transformReply ?
                transformReply(reply, args.preserve) :
                reply;
        };
    }
    static _createModuleCommand(command, resp) {
        const transformReply = (0, commander_1.getTransformReply)(command, resp);
        return async function () {
            const args = command.transformArguments.apply(undefined, arguments), reply = await this.self._sendCommand(args, this.self.commandOptions);
            return transformReply ?
                transformReply(reply, args.preserve) :
                reply;
        };
    }
    static _createFunctionCommand(name, fn, resp) {
        const prefix = (0, commander_1.functionArgumentsPrefix)(name, fn), transformReply = (0, commander_1.getTransformReply)(fn, resp);
        return async function () {
            const fnArgs = fn.transformArguments.apply(undefined, arguments), reply = await this.self._sendCommand(prefix.concat(fnArgs), this.self.commandOptions);
            return transformReply ?
                transformReply(reply, fnArgs.preserve) :
                reply;
        };
    }
    static _createScriptCommand(script, resp) {
        const prefix = (0, commander_1.scriptArgumentsPrefix)(script), transformReply = (0, commander_1.getTransformReply)(script, resp);
        return async function () {
            const scriptArgs = script.transformArguments.apply(undefined, arguments), args = prefix.concat(scriptArgs), reply = await this._sendCommand(args, this.commandOptions).catch((err) => {
                if (!err?.message?.startsWith?.('NOSCRIPT'))
                    throw err;
                args[0] = 'EVAL';
                args[1] = script.SCRIPT;
                return this._sendCommand(args, this.commandOptions);
            });
            return transformReply ?
                transformReply(reply, scriptArgs.preserve) :
                reply;
        };
    }
    static factory(config) {
        const Client = (0, commander_1.attachConfig)({
            BaseClass: RedisClient,
            commands: commands_1.default,
            createCommand: RedisClient._createCommand,
            createFunctionCommand: RedisClient._createFunctionCommand,
            createModuleCommand: RedisClient._createModuleCommand,
            createScriptCommand: RedisClient._createScriptCommand,
            config
        });
        Client.prototype.Multi = multi_command_1.default.extend(config);
        return (options) => {
            // returning a proxy of the client to prevent the namespaces.self to leak between proxies
            // namespaces will be bootstraped on first access per proxy
            return Object.create(new Client(options));
        };
    }
    static create(options) {
        return RedisClient.factory(options)(options);
    }
    static parseURL(url) {
        // https://www.iana.org/assignments/uri-schemes/prov/redis
        const { hostname, port, protocol, username, password, pathname } = new url_1.URL(url), parsed = {
            socket: {
                host: hostname
            }
        };
        if (protocol === 'rediss:') {
            parsed.socket.tls = true;
        }
        else if (protocol !== 'redis:') {
            throw new TypeError('Invalid protocol');
        }
        if (port) {
            parsed.socket.port = Number(port);
        }
        if (username) {
            parsed.username = decodeURIComponent(username);
        }
        if (password) {
            parsed.password = decodeURIComponent(password);
        }
        if (pathname.length > 1) {
            const database = Number(pathname.substring(1));
            if (isNaN(database)) {
                throw new TypeError('Invalid pathname');
            }
            parsed.database = database;
        }
        return parsed;
    }
    get options() {
        return this._options;
    }
    get isOpen() {
        return this._socket.isOpen;
    }
    get isReady() {
        return this._socket.isReady;
    }
    get isPubSubActive() {
        return this._queue.isPubSubActive;
    }
    // get v4(): Record<string, any> {
    //   if (!this.client.#options?.legacyMode) {
    //     throw new Error('the client is not in "legacy mode"');
    //   }
    //   return this.client.#v4;
    // }
    constructor(options) {
        super();
        Object.defineProperty(this, "self", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this
        });
        Object.defineProperty(this, "_options", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_socket", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_queue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_isolationPool", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        // readonly #v4: Record<string, any> = {};
        Object.defineProperty(this, "_selectedDB", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 0
        });
        // #legacyMode(): void {
        //   if (!this.#options?.legacyMode) return;
        //   (this as any).#v4.sendCommand = this.#sendCommand.bind(this);
        //   (this as any).sendCommand = (...args: Array<any>): void => {
        //     const result = this.#legacySendCommand(...args);
        //     if (result) {
        //       result.promise
        //         .then(reply => result.callback(null, reply))
        //         .catch(err => result.callback(err));
        //     }
        //   };
        //   for (const [name, command] of Object.entries(COMMANDS)) {
        //     this.#defineLegacyCommand(name, command);
        //     (this as any)[name.toLowerCase()] ??= (this as any)[name];
        //   }
        //   // hard coded commands
        //   this.#defineLegacyCommand('SELECT');
        //   this.#defineLegacyCommand('select');
        //   this.#defineLegacyCommand('SUBSCRIBE');
        //   this.#defineLegacyCommand('subscribe');
        //   this.#defineLegacyCommand('PSUBSCRIBE');
        //   this.#defineLegacyCommand('pSubscribe');
        //   this.#defineLegacyCommand('UNSUBSCRIBE');
        //   this.#defineLegacyCommand('unsubscribe');
        //   this.#defineLegacyCommand('PUNSUBSCRIBE');
        //   this.#defineLegacyCommand('pUnsubscribe');
        //   this.#defineLegacyCommand('QUIT');
        //   this.#defineLegacyCommand('quit');
        // }
        // #legacySendCommand(...args: Array<any>) {
        //   const callback = typeof args[args.length - 1] === 'function' ?
        //     args.pop() as ClientLegacyCallback :
        //     undefined;
        //   const promise = this.#sendCommand(transformLegacyCommandArguments(args));
        //   if (callback) return {
        //     promise,
        //     callback
        //   };
        //   promise.catch(err => this.emit('error', err));
        // }
        // #defineLegacyCommand(name: string, command?: RedisCommand): void {
        //   this.#v4[name] = (this as any)[name].bind(this);
        //   (this as any)[name] = command && command.TRANSFORM_LEGACY_REPLY && command.transformReply ?
        //     (...args: Array<unknown>) => {
        //       const result = this.#legacySendCommand(name, ...args);
        //       if (result) {
        //         result.promise
        //           .then(reply => result.callback(null, command.transformReply!(reply)))
        //           .catch(err => result.callback(err));
        //       }
        //     } :
        //     (...args: Array<unknown>) => (this as any).sendCommand(name, ...args);
        // }
        Object.defineProperty(this, "_pingTimer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "sendCommand", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this._sendCommand.bind(this)
        });
        Object.defineProperty(this, "select", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SELECT
        });
        Object.defineProperty(this, "subscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SUBSCRIBE
        });
        Object.defineProperty(this, "unsubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.UNSUBSCRIBE
        });
        Object.defineProperty(this, "pSubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.PSUBSCRIBE
        });
        Object.defineProperty(this, "pUnsubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.PUNSUBSCRIBE
        });
        Object.defineProperty(this, "sSubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SSUBSCRIBE
        });
        Object.defineProperty(this, "sUnsubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SUNSUBSCRIBE
        });
        Object.defineProperty(this, "quit", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.QUIT
        });
        Object.defineProperty(this, "multi", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.MULTI
        });
        this._options = this._initiateOptions(options);
        this._queue = this._initiateQueue();
        this._socket = this._initiateSocket();
        // this.#legacyMode();
    }
    _initiateOptions(options) {
        if (options?.url) {
            const parsed = RedisClient.parseURL(options.url);
            if (options.socket) {
                parsed.socket = Object.assign(options.socket, parsed.socket);
            }
            Object.assign(options, parsed);
        }
        if (options?.database) {
            this._selectedDB = options.database;
        }
        return options;
    }
    _initiateQueue() {
        return new commands_queue_1.default(this._options?.RESP, this._options?.commandsQueueMaxLength, (channel, listeners) => this.emit('sharded-channel-moved', channel, listeners));
    }
    _initiateSocket() {
        const socketInitiator = async () => {
            const promises = [];
            if (this._selectedDB !== 0) {
                promises.push(this._queue.addCommand(['SELECT', this._selectedDB.toString()], { asap: true }));
            }
            if (this._options?.readonly) {
                // promises.push(
                //     this.#queue.addCommand(
                //         COMMANDS.READONLY.transformArguments(),
                //         { asap: true }
                //     )
                // );
            }
            if (this._options?.RESP) {
                const hello = {};
                if (this._options.password) {
                    hello.AUTH = {
                        username: this._options.username ?? 'default',
                        password: this._options.password
                    };
                }
                if (this._options.name) {
                    hello.SETNAME = this._options.name;
                }
                promises.push(this._queue.addCommand(HELLO_1.default.transformArguments(this._options.RESP, hello), { asap: true }));
            }
            else {
                if (this._options?.name) {
                    // promises.push(
                    //     this.#queue.addCommand(
                    //         COMMANDS.CLIENT_SETNAME.transformArguments(this.#options.name),
                    //         { asap: true }
                    //     )
                    // );
                }
                if (this._options?.username || this._options?.password) {
                    // promises.push(
                    //     this.#queue.addCommand(
                    //         COMMANDS.AUTH.transformArguments({
                    //             username: this.#options.username,
                    //             password: this.#options.password ?? ''
                    //         }),
                    //         { asap: true }
                    //     )
                    // );
                }
            }
            const resubscribePromise = this._queue.resubscribe();
            if (resubscribePromise) {
                promises.push(resubscribePromise);
            }
            if (promises.length) {
                this._tick(true);
                await Promise.all(promises);
            }
        };
        return new socket_1.default(socketInitiator, this._options?.socket)
            .on('data', chunk => this._queue.decoder.write(chunk))
            .on('error', err => {
            this.emit('error', err);
            if (this._socket.isOpen && !this._options?.disableOfflineQueue) {
                this._queue.flushWaitingForReply(err);
            }
            else {
                this._queue.flushAll(err);
            }
        })
            .on('connect', () => this.emit('connect'))
            .on('ready', () => {
            this.emit('ready');
            this._setPingTimer();
            this._tick();
        })
            .on('reconnecting', () => this.emit('reconnecting'))
            .on('drain', () => this._tick())
            .on('end', () => this.emit('end'));
    }
    _setPingTimer() {
        if (!this._options?.pingInterval || !this._socket.isReady)
            return;
        clearTimeout(this._pingTimer);
        this._pingTimer = setTimeout(() => {
            if (!this._socket.isReady)
                return;
            // using _sendCommand to support legacy mode
            this._sendCommand(['PING'])
                .then(reply => this.emit('ping-interval', reply))
                .catch(err => this.emit('error', err))
                .finally(() => this._setPingTimer());
        }, this._options.pingInterval);
    }
    withCommandOptions(options) {
        const proxy = Object.create(this.self);
        proxy.commandOptions = options;
        return proxy;
    }
    _commandOptionsProxy(key, value) {
        const proxy = Object.create(this.self);
        proxy.commandOptions = Object.create(this.commandOptions ?? null);
        proxy.commandOptions[key] = value;
        return proxy;
    }
    /**
     * Override the `flags` command option
     */
    withFlags(flags) {
        return this._commandOptionsProxy('flags', flags);
    }
    /**
     * Override the `asap` command option to `true`
     */
    asap() {
        return this._commandOptionsProxy('asap', true);
    }
    /**
     * Override the `isolated` command option to `true`
     */
    isolated() {
        return this._commandOptionsProxy('isolated', true);
    }
    duplicate(overrides) {
        return new (Object.getPrototypeOf(this).constructor)({
            ...this._options,
            ...overrides
        });
    }
    async connect() {
        await this._socket.connect();
        this.self._isolationPool = (0, generic_pool_1.createPool)({
            create: async () => {
                const duplicate = this.duplicate({
                    isolationPoolOptions: undefined
                }).on('error', err => this.emit('error', err));
                await duplicate.connect();
                return duplicate;
            },
            destroy: client => client.disconnect()
        }, this._options?.isolationPoolOptions);
    }
    // using `_` to avoid conflicts with the legacy mode
    _sendCommand(args, options) {
        if (!this._socket.isOpen) {
            return Promise.reject(new errors_1.ClientClosedError());
        }
        else if (options?.isolated) {
            return this.executeIsolated(isolatedClient => isolatedClient.sendCommand(args, {
                ...options,
                isolated: false
            }));
        }
        else if (!this._socket.isReady && this._options?.disableOfflineQueue) {
            return Promise.reject(new errors_1.ClientOfflineError());
        }
        const promise = this._queue.addCommand(args, options);
        this._tick();
        return promise;
    }
    async SELECT(db) {
        await this._sendCommand(['SELECT', db.toString()]);
        this._selectedDB = db;
    }
    _pubSubCommand(promise) {
        if (promise === undefined)
            return Promise.resolve();
        this._tick();
        return promise;
    }
    SUBSCRIBE(channels, listener, bufferMode) {
        console.log('SUBSCRIBE', channels, listener, bufferMode, this._options?.RESP);
        return this._pubSubCommand(this._queue.subscribe(pub_sub_1.PubSubType.CHANNELS, channels, listener, bufferMode));
    }
    UNSUBSCRIBE(channels, listener, bufferMode) {
        return this._pubSubCommand(this._queue.unsubscribe(pub_sub_1.PubSubType.CHANNELS, channels, listener, bufferMode));
    }
    PSUBSCRIBE(patterns, listener, bufferMode) {
        return this._pubSubCommand(this._queue.subscribe(pub_sub_1.PubSubType.PATTERNS, patterns, listener, bufferMode));
    }
    PUNSUBSCRIBE(patterns, listener, bufferMode) {
        return this._pubSubCommand(this._queue.unsubscribe(pub_sub_1.PubSubType.PATTERNS, patterns, listener, bufferMode));
    }
    SSUBSCRIBE(channels, listener, bufferMode) {
        return this._pubSubCommand(this._queue.subscribe(pub_sub_1.PubSubType.SHARDED, channels, listener, bufferMode));
    }
    SUNSUBSCRIBE(channels, listener, bufferMode) {
        return this._pubSubCommand(this._queue.unsubscribe(pub_sub_1.PubSubType.SHARDED, channels, listener, bufferMode));
    }
    getPubSubListeners(type) {
        return this._queue.getPubSubListeners(type);
    }
    extendPubSubChannelListeners(type, channel, listeners) {
        return this._pubSubCommand(this._queue.extendPubSubChannelListeners(type, channel, listeners));
    }
    extendPubSubListeners(type, listeners) {
        return this._pubSubCommand(this._queue.extendPubSubListeners(type, listeners));
    }
    QUIT() {
        return this._socket.quit(async () => {
            const quitPromise = this._queue.addCommand(['QUIT']);
            this._tick();
            const [reply] = await Promise.all([
                quitPromise,
                this._destroyIsolationPool()
            ]);
            return reply;
        });
    }
    _tick(force = false) {
        if (this._socket.writableNeedDrain || (!force && !this._socket.isReady)) {
            return;
        }
        this._socket.cork();
        while (!this._socket.writableNeedDrain) {
            const args = this._queue.getCommandToSend();
            if (args === undefined)
                break;
            this._socket.writeCommand(args);
        }
    }
    executeIsolated(fn) {
        return this._isolationPool ?
            this._isolationPool.use(fn) :
            Promise.reject(new errors_1.ClientClosedError());
    }
    _addMultiCommands(commands, chainId, flags) {
        return Promise.all(commands.map(({ args }) => this._queue.addCommand(args, {
            chainId,
            flags
        })));
    }
    MULTI() {
        return new this.Multi(async (commands, selectedDB, chainId) => {
            if (!this._socket.isOpen) {
                return Promise.reject(new errors_1.ClientClosedError());
            }
            const flags = this.commandOptions?.flags, promise = chainId ?
                // if `chainId` has a value, it's a `MULTI` (and not "pipeline") - need to add the `MULTI` and `EXEC` commands
                Promise.all([
                    this._queue.addCommand(['MULTI'], { chainId }),
                    this._addMultiCommands(commands, chainId),
                    this._queue.addCommand(['EXEC'], { chainId, flags })
                ]) :
                this._addMultiCommands(commands, undefined, flags);
            this._tick();
            const results = await promise;
            if (selectedDB !== undefined) {
                this._selectedDB = selectedDB;
            }
            return results;
        }
        // self.#options?.legacyMode
        );
    }
    async *scanIterator(options) {
        let cursor = options?.cursor ?? 0;
        do {
            const reply = await this.scan(cursor, options);
            cursor = reply.cursor;
            yield reply.keys;
        } while (cursor !== 0);
    }
    async *hScanIterator(key, options) {
        let cursor = options?.cursor ?? 0;
        do {
            const reply = await this.hScan(key, cursor, options);
            cursor = reply.cursor;
            yield reply.entries;
        } while (cursor !== 0);
    }
    async *sScanIterator(key, options) {
        let cursor = options?.cursor ?? 0;
        do {
            const reply = await this.sScan(key, cursor, options);
            cursor = reply.cursor;
            yield reply.members;
        } while (cursor !== 0);
    }
    async *zScanIterator(key, options) {
        let cursor = options?.cursor ?? 0;
        do {
            const reply = await this.zScan(key, cursor, options);
            cursor = reply.cursor;
            yield reply.members;
        } while (cursor !== 0);
    }
    async disconnect() {
        this._queue.flushAll(new errors_1.DisconnectsClientError());
        this._socket.disconnect();
        await this._destroyIsolationPool();
    }
    async _destroyIsolationPool() {
        await this._isolationPool.drain();
        await this._isolationPool.clear();
        this.self._isolationPool = undefined;
    }
    ref() {
        this._socket.ref();
    }
    unref() {
        this._socket.unref();
    }
}
exports.default = RedisClient;
