"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const commands_1 = require("../commands");
const events_1 = require("events");
const commander_1 = require("../commander");
const cluster_slots_1 = require("./cluster-slots");
const errors_1 = require("../errors");
class RedisCluster extends events_1.EventEmitter {
    static _extractFirstKey(command, args, redisArgs) {
        if (command.FIRST_KEY_INDEX === undefined) {
            return undefined;
        }
        else if (typeof command.FIRST_KEY_INDEX === 'number') {
            return redisArgs[command.FIRST_KEY_INDEX];
        }
        return command.FIRST_KEY_INDEX(...args);
    }
    static _createCommand(command, resp) {
        const transformReply = (0, commander_1.getTransformReply)(command, resp);
        return async function () {
            const args = command.transformArguments.apply(undefined, arguments), firstKey = RedisCluster._extractFirstKey(command, arguments, args), reply = await this.sendCommand(firstKey, command.IS_READ_ONLY, args, this.commandOptions, command.POLICIES);
            return transformReply ?
                transformReply(reply, args.preserve) :
                reply;
        };
    }
    static _createModuleCommand(command, resp) {
        const transformReply = (0, commander_1.getTransformReply)(command, resp);
        return async function () {
            const args = command.transformArguments.apply(undefined, arguments), firstKey = RedisCluster._extractFirstKey(command, arguments, args), reply = await this.self.sendCommand(firstKey, command.IS_READ_ONLY, args, this.self.commandOptions, command.POLICIES);
            return transformReply ?
                transformReply(reply, args.preserve) :
                reply;
        };
    }
    static _createFunctionCommand(name, fn, resp) {
        const prefix = (0, commander_1.functionArgumentsPrefix)(name, fn), transformReply = (0, commander_1.getTransformReply)(fn, resp);
        return async function () {
            const fnArgs = fn.transformArguments.apply(undefined, arguments), args = prefix.concat(fnArgs), firstKey = RedisCluster._extractFirstKey(fn, arguments, args), reply = await this.self.sendCommand(firstKey, fn.IS_READ_ONLY, args, this.self.commandOptions, fn.POLICIES);
            return transformReply ?
                transformReply(reply, fnArgs.preserve) :
                reply;
        };
    }
    static _createScriptCommand(script, resp) {
        const prefix = (0, commander_1.scriptArgumentsPrefix)(script), transformReply = (0, commander_1.getTransformReply)(script, resp);
        return async function () {
            const scriptArgs = script.transformArguments.apply(undefined, arguments), args = prefix.concat(scriptArgs), firstKey = RedisCluster._extractFirstKey(script, arguments, args), reply = await this.sendCommand(firstKey, script.IS_READ_ONLY, args, this.commandOptions, script.POLICIES);
            return transformReply ?
                transformReply(reply, scriptArgs.preserve) :
                reply;
        };
    }
    static factory(config) {
        const Cluster = (0, commander_1.attachConfig)({
            BaseClass: RedisCluster,
            commands: commands_1.default,
            createCommand: RedisCluster._createCommand,
            createFunctionCommand: RedisCluster._createFunctionCommand,
            createModuleCommand: RedisCluster._createModuleCommand,
            createScriptCommand: RedisCluster._createScriptCommand,
            config
        });
        // Client.prototype.Multi = RedisClientMultiCommand.extend(config);
        return (options) => {
            // returning a proxy of the client to prevent the namespaces.self to leak between proxies
            // namespaces will be bootstraped on first access per proxy
            return Object.create(new Cluster(options));
        };
    }
    static create(options) {
        return RedisCluster.factory(options)(options);
    }
    /**
     * An array of the cluster slots, each slot contain its `master` and `replicas`.
     * Use with {@link RedisCluster.prototype.nodeClient} to get the client for a specific node (master or replica).
     */
    get slots() {
        return this._slots.slots;
    }
    /**
     * An array of cluster shards, each shard contain its `master` and `replicas`.
     * Use with {@link RedisCluster.prototype.nodeClient} to get the client for a specific node (master or replica).
     */
    get shards() {
        return this._slots.shards;
    }
    /**
     * An array of the cluster masters.
     * Use with {@link RedisCluster.prototype.nodeClient} to get the client for a specific master node.
     */
    get masters() {
        return this._slots.masters;
    }
    /**
     * An array of the cluster replicas.
     * Use with {@link RedisCluster.prototype.nodeClient} to get the client for a specific replica node.
     */
    get replicas() {
        return this._slots.replicas;
    }
    /**
     * A map form a node address (`<host>:<port>`) to its shard, each shard contain its `master` and `replicas`.
     * Use with {@link RedisCluster.prototype.nodeClient} to get the client for a specific node (master or replica).
     */
    get nodeByAddress() {
        return this._slots.nodeByAddress;
    }
    /**
     * The current pub/sub node.
     */
    get pubSubNode() {
        return this._slots.pubSubNode;
    }
    // readonly #Multi: InstantiableRedisClusterMultiCommandType<M, F, S>;
    get isOpen() {
        return this._slots.isOpen;
    }
    constructor(options) {
        super();
        Object.defineProperty(this, "_options", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "_slots", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "subscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SUBSCRIBE
        });
        Object.defineProperty(this, "unsubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.UNSUBSCRIBE
        });
        Object.defineProperty(this, "pSubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.PSUBSCRIBE
        });
        Object.defineProperty(this, "pUnsubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.PUNSUBSCRIBE
        });
        Object.defineProperty(this, "sSubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SSUBSCRIBE
        });
        Object.defineProperty(this, "sUnsubscribe", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: this.SUNSUBSCRIBE
        });
        this._options = options;
        this._slots = new cluster_slots_1.default(options, this.emit.bind(this));
        // this.#Multi = RedisClusterMultiCommand.extend(options);
    }
    duplicate(overrides) {
        return new (Object.getPrototypeOf(this).constructor)({
            ...this._options,
            ...overrides
        });
    }
    connect() {
        return this._slots.connect();
    }
    withCommandOptions(options) {
        const proxy = Object.create(this);
        proxy.commandOptions = options;
        return proxy;
    }
    _commandOptionsProxy(key, value) {
        const proxy = Object.create(this);
        proxy.commandOptions = Object.create(this.commandOptions ?? null);
        proxy.commandOptions[key] = value;
        return proxy;
    }
    /**
     * Override the `flags` command option
     */
    withFlags(flags) {
        return this._commandOptionsProxy('flags', flags);
    }
    /**
     * Override the `policies` command option
     * TODO
     */
    withPolicies(policies) {
        return this._commandOptionsProxy('policies', policies);
    }
    async sendCommand(firstKey, isReadonly, args, options, deafultPolicies) {
        // const requestPolicy = options?.policies?.request ?? deafultPolicies?.request,
        //   responsePolicy = options?.policies?.response ?? deafultPolicies?.response;
        const maxCommandRedirections = this._options.maxCommandRedirections ?? 16;
        let client = await this._slots.getClient(firstKey, isReadonly);
        for (let i = 0;; i++) {
            try {
                return await client.sendCommand(args, options);
            }
            catch (err) {
                // TODO: error class
                if (++i > maxCommandRedirections || !(err instanceof Error)) {
                    throw err;
                }
                if (err.message.startsWith('ASK')) {
                    const address = err.message.substring(err.message.lastIndexOf(' ') + 1);
                    let redirectTo = await this._slots.getMasterByAddress(address);
                    if (!redirectTo) {
                        await this._slots.rediscover(client);
                        redirectTo = await this._slots.getMasterByAddress(address);
                    }
                    if (!redirectTo) {
                        throw new Error(`Cannot find node ${address}`);
                    }
                    await redirectTo.asking();
                    client = redirectTo;
                    continue;
                }
                else if (err.message.startsWith('MOVED')) {
                    await this._slots.rediscover(client);
                    client = await this._slots.getClient(firstKey, isReadonly);
                    continue;
                }
                throw err;
            }
        }
    }
    // MULTI(routing?: RedisCommandArgument): RedisClusterMultiCommandType<M, F, S> {
    //   return new this.#Multi(
    //     (commands: Array<RedisMultiQueuedCommand>, firstKey?: RedisCommandArgument, chainId?: symbol) => {
    //       return this.#execute(
    //         firstKey,
    //         false,
    //         client => client.multiExecutor(commands, undefined, chainId)
    //       );
    //     },
    //     routing
    //   );
    // }
    // multi = this.MULTI;
    async SUBSCRIBE(channels, listener, bufferMode) {
        return (await this._slots.getPubSubClient())
            .SUBSCRIBE(channels, listener, bufferMode);
    }
    async UNSUBSCRIBE(channels, listener, bufferMode) {
        return this._slots.executeUnsubscribeCommand(client => client.UNSUBSCRIBE(channels, listener, bufferMode));
    }
    async PSUBSCRIBE(patterns, listener, bufferMode) {
        return (await this._slots.getPubSubClient())
            .PSUBSCRIBE(patterns, listener, bufferMode);
    }
    async PUNSUBSCRIBE(patterns, listener, bufferMode) {
        return this._slots.executeUnsubscribeCommand(client => client.PUNSUBSCRIBE(patterns, listener, bufferMode));
    }
    async SSUBSCRIBE(channels, listener, bufferMode) {
        const maxCommandRedirections = this._options.maxCommandRedirections ?? 16, firstChannel = Array.isArray(channels) ? channels[0] : channels;
        let client = await this._slots.getShardedPubSubClient(firstChannel);
        for (let i = 0;; i++) {
            try {
                return await client.SSUBSCRIBE(channels, listener, bufferMode);
            }
            catch (err) {
                if (++i > maxCommandRedirections || !(err instanceof errors_1.ErrorReply)) {
                    throw err;
                }
                if (err.message.startsWith('MOVED')) {
                    await this._slots.rediscover(client);
                    client = await this._slots.getShardedPubSubClient(firstChannel);
                    continue;
                }
                throw err;
            }
        }
    }
    SUNSUBSCRIBE(channels, listener, bufferMode) {
        return this._slots.executeShardedUnsubscribeCommand(Array.isArray(channels) ? channels[0] : channels, client => client.SUNSUBSCRIBE(channels, listener, bufferMode));
    }
    // quit(): Promise<void> {
    //   return this.#slots.quit();
    // }
    disconnect() {
        return this._slots.disconnect();
    }
    nodeClient(node) {
        return this._slots.nodeClient(node);
    }
    /**
     * Returns a random node from the cluster.
     * Userful for running "forward" commands (like PUBLISH) on a random node.
     */
    getRandomNode() {
        return this._slots.getRandomNode();
    }
    /**
     * Get a random node from a slot.
     * Useful for running readonly commands on a slot.
     */
    getSlotRandomNode(slot) {
        return this._slots.getSlotRandomNode(slot);
    }
    /**
     * @deprecated use `.masters` instead
     * TODO
     */
    getMasters() {
        return this.masters;
    }
    /**
     * @deprecated use `.slots[<SLOT>]` instead
     * TODO
     */
    getSlotMaster(slot) {
        return this.slots[slot].master;
    }
}
exports.default = RedisCluster;
